@@ -1,9 +1,8 @@
-#!/usr/bin/env perl
 package Mouse::Object;
 use strict;
 use warnings;
 
-use Mouse::Util qw/weaken/;
+use Scalar::Util 'weaken';
 use Carp 'confess';
 
 sub new {
@@ -16,11 +15,11 @@ sub new {
     for my $attribute ($class->meta->compute_all_applicable_attributes) {
         my $from = $attribute->init_arg;
         my $key  = $attribute->name;
-        my $default;
 
         if (defined($from) && exists($args->{$from})) {
-            $attribute->verify_type_constraint($args->{$from})
-                if $attribute->has_type_constraint;
+            $args->{$from} = $attribute->coerce_constraint($args->{$from})
+                if $attribute->should_coerce;
+            $attribute->verify_against_type_constraint($args->{$from});
 
             $instance->{$key} = $args->{$from};
 
@@ -39,11 +38,12 @@ sub new {
                     my $value = $attribute->has_builder
                               ? $instance->$builder
                               : ref($default) eq 'CODE'
-                                  ? $default->()
+                                  ? $default->($instance)
                                   : $default;
 
-                    $attribute->verify_type_constraint($value)
-                        if $attribute->has_type_constraint;
+                    $value = $attribute->coerce_constraint($value)
+                        if $attribute->should_coerce;
+                    $attribute->verify_against_type_constraint($value);
 
                     $instance->{$key} = $value;
 
